// Distributes Workspace Views structures to other users in your tenant
// Author: Lee Worton, Autodesk
// Last Modification Date : Mar 16, 2020
// Usage : node copyViews

/* ----------------------------------------------------
     Do not modify below
   ---------------------------------------------------- */

let userViews   = [];
let tableaus    = [];
let index       = 0;
let optionsFile = "options.js";

if(process.argv.length > 2) {  optionsFile = process.argv[2]; }

const axios         = require('axios');
const querystring   = require('querystring');
const settings      = require("./settings.js");
const options       = require("./" + optionsFile);


console.log("   ");
console.log("   *********************************** START ***********************************");
console.log("   ");
console.log("    SETTINGS");
console.log("   -----------------------------------------------------------------------------");
console.log("    Tenant           : " + settings.tenant); 
console.log("    User Name        : " + settings.user); 
console.log("    Copy to          : " + options.distributeTo); 
console.log("    Workspaces count : " + options.viewsToCopy.length); 
console.log("   "); 


login(function() {
    getUsers(function() {
        
        console.log();
        console.log();
        console.log("    GETTING DEFINED VIEWS");
        console.log("   -----------------------------------------------------------------------------");
        
        getWorkspacesViews();
        
    })
});


// Login to FORGE
function login(callback) {
    
    let params = {
        "client_id"     : settings.clientId,
        "client_secret" : settings.clientSecret,
        "grant_type"    : 'client_credentials',
        "scope"         : "data:read"
    }
    
    axios.post('https://developer.api.autodesk.com/authentication/v1/authenticate', querystring.stringify(params)).then(function (response) {
        if (response.status == 200) {               

            axios.defaults.headers.common['Content-Type']   = "application/json";
            axios.defaults.headers.common['Accept']         = "";
            axios.defaults.headers.common['X-user-id']      = settings.user;
            axios.defaults.headers.common['X-Tenant']       = settings.tenant;
            axios.defaults.headers.common['Authorization']  = "Bearer " + response.data.access_token;
            
            console.log("    > Login to tenant " + settings.tenant + " successful");
            
            callback();

        } else {

            console.log("");      
            console.log("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");      
            console.log("             LOGIN FAILED");
            console.log("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"); 
            console.log(""); 
            
            console.log(error);
        }

  }).catch(function (error) {
    console.log(error);
  });
    
}


// Get all active users except for the main user
function getUsers(callback) {
    
    let url     = "https://" + settings.tenant + ".autodeskplm360.net/api/v3/users?limit=1000";
    let groups  = options.distributeTo.split(",");
    
    axios.get(url).then(function (response) {
        
        console.log("    > Found " + response.data.items.length + " active users in total");

        for(user of response.data.items) {
            if (user.email != settings.user) {
                
                let add = (options.distributeTo === "*");    
                
                for(group of user.groups) {
                    if(groups.includes(group.shortName)) {
                        add = true;
                        false;
                   }
                }
                
                if(add) userViews.push({
                    email        : user.email,
                    views        : [],
                    currentId    : "",
                    currentViews : []
                });
                
            }
        }
        
        console.log("    > Copying view(s) to " + userViews.length + " users");
        
        callback();
   
    }).catch(function (error) {
        console.log(error.message);
    });

}


// Get workspaces views defined for current user
function getWorkspacesViews() {
    
    if(index < options.viewsToCopy.length) {
        
        getWorkspaceViews(function() {
            index++;
            getWorkspacesViews();
        });
        
    } else {
        
        console.log("    Found " + tableaus.length + " views in total to copy");
        console.log();
        
        extractWorkspaceViews(function() {
            copyViews();
        });
        
    }
    
}
function getWorkspaceViews(callback) {    
    
    let url       = "https://" + settings.tenant + ".autodeskplm360.net/api/v3/workspaces/" + options.viewsToCopy[index].wsId + "/tableaus";
    let userViews = options.viewsToCopy[index].views.split(",");
    
    axios.get(url,  {
        headers : {
            'Accept' : "application/json"
        }
    }).then(function (response) {
        
	    for (i = 0; i < userViews.length; i++) {
            
		    for(tableau of response.data.tableaus) {
				if (tableau.title === userViews[i].trim()) {

                    console.log("    WS " + options.viewsToCopy[index].wsId + " : Found the view '" + tableau.title + "'");
                    
                    if(options.viewsToCopy[index].hasOwnProperty("default") && options.viewsToCopy[index].default === tableau.title) {
                        tableau.setCopyDefault = true;
                    } else {
                        tableau.setCopyDefault = false;
                    }
                    
                    if(options.viewsToCopy[index].hasOwnProperty("force") && options.viewsToCopy[index].force === true) {
                        tableau.force = true;
                    } else {
                        tableau.force = false;
                    }
                    
                    tableaus.push(tableau);
                    break;
				}
			}
            
		}
        
        callback();
   
    }).catch(function (error) {
        console.log();
        console.log();
        console.log("   -----------------------------------------------------------------------------");
        console.log("    ERROR in requesting " + url);
        console.log("   -----------------------------------------------------------------------------");
        console.log();
    });
    
}


// Get configuration of views to copy
function extractWorkspaceViews(callback) {
     
    console.log();
    console.log("    GETTING VIEW CONFIGURATIONS");
    console.log("   -----------------------------------------------------------------------------");
    
    axios.defaults.headers.common['Accept'] = "application/vnd.autodesk.plm.meta+json";
    
    let requests = [];
    let configs  = [];
    
    for(tableau of tableaus) {
        console.log("    Requesting " + tableau.link);
        requests.push(extractWorkspaceView(configs, tableau));
    }
    
    axios.all(requests).then(axios.spread(function (acct, perms) {
        
        console.log();
        console.log();
        console.log("    COPYING VIEWS");
        console.log("   -----------------------------------------------------------------------------");
        
        for(user of userViews) user.views = JSON.parse(JSON.stringify(configs));;
        
        callback();
        
    }));
    
}
function extractWorkspaceView(configs, tableau) {
    
    return axios.get("https://" + settings.tenant + ".autodeskplm360.net/" + tableau.link).then(function (response) {
        
        let config = response.data;
            config.setCopyDefault = tableau.setCopyDefault;
            config.force = tableau.force;
        
        configs.push(config);
    }).catch(function (error) {
        console.log(error);
    });
         
}


// Copy views to users
function copyViews() {
    
    if(userViews.length === 0) {
        console.log("   ");
        console.log("   ************************************ END ************************************");
        console.log("   ");
    } else {
        copyView();
    }
    
}
function copyView() {
        
    let userView = userViews[0];
    
    if(userView.views.length === 0) {
        
        userViews.splice(0, 1);
        copyViews();
        
    } else {

        let config      = userView.views[0];
        let currentId   = config.workspace.link.split("/")[4];
        let url         = "https://" + settings.tenant + ".autodeskplm360.net" + config.workspace.link + "/tableaus";
        
        if(userView.currentId !== currentId) { 

            userView.currentId = currentId;
            getCurrentViews(userView, currentId, url);
            
        } else {
        
            let mode = "create";
            
            for(tableau of userView.currentViews.tableaus) {
                if(tableau.title === config.name) {
                    url  = "https://" + settings.tenant + ".autodeskplm360.net" + tableau.link;
                    mode = (config.force === true) ? "update" : "skip";
                    break;
                }
            }
        
                 if(mode === "create") { createView(userView, currentId, config, url); }
            else if(mode === "update") { updateView(userView, currentId, config, url); }
            else {
                console.log("    WS " + currentId + " : Skipping view '" + config.name + "' for user " + userView.email);
                userView.views.splice(0,1);
                copyViews();
            }
            
        }
        
    }
    
}
function getCurrentViews(userView, currentId, url) {
    
    console.log("    WS " + currentId + " : Getting current views of user " + userView.email);
    
     axios.get(url,{
            headers : {
                'Accept'        : "application/json",
                'X-user-id'     : userView.email
            }
        }).then(function (response) {
         
            userView.currentViews = response.data;
            copyViews();
         
    }).catch(function (error) {
            
        console.log();
        console.log("   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        console.log("    ERROR in getting current views:");
        console.log(error.response.data);
        console.log("   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        console.log();
            
    });
    
}
function createView(userView, currentId, config, url) {
 
    console.log("    WS " + currentId + " : Copying '" + config.name + "' to " + userView.email + " (default : " + config.setCopyDefault + ")");
    
    axios.post(url,{
        name                    : config.name,
        columns                 : config.columns,
        description             : config.description,
        showOnlyDeletedRecords  : config.showOnlyDeletedRecords,
        isDefault               : config.setCopyDefault,
        createdDate             : new Date()
    },{
        headers : {
            'Accept'        : "application/json",
            'Content-Type'  : "application/vnd.autodesk.plm.meta+json",
            'X-user-id'     : userView.email
        }
    }).then(function (response) {

        if(response.status !== 201) {

            console.log();
            console.log("   ---------------------------------------------------------------");
            console.log("    ERROR in creating view");
            console.log("   ---------------------------------------------------------------");
            console.log();

        } else {

            userView.views.splice(0,1);
            copyViews();

        }

    }).catch(function (error) {

        console.log();
        console.log("   ---------------------------------------------------------------");
        console.log("    ERROR in creating view");
        console.log("   ---------------------------------------------------------------");
        console.log();
        console.log(error);
        console.log();
        console.log();

    });
    
}
function updateView(userView, currentId, config, url) {
 
    console.log("    WS " + currentId + " : Updating '" + config.name + "' for " + userView.email + " (default : " + config.setCopyDefault + ")");
    
    axios.put(url,{
        name                    : config.name,
        columns                 : config.columns,
        description             : config.description,
        showOnlyDeletedRecords  : config.showOnlyDeletedRecords,
        isDefault               : config.setCopyDefault,
        createdDate             : new Date()
    },{
        headers : {
            'Accept'        : "application/json",
            'Content-Type'  : "application/vnd.autodesk.plm.meta+json",
            'X-user-id'     : userView.email
        }
    }).then(function (response) {

        
        if(response.status !== 204) {

            console.log();
            console.log("   ---------------------------------------------------------------");
            console.log("    ERROR in updating view");
            console.log("   ---------------------------------------------------------------");
            console.log();

        } else {

            userView.views.splice(0,1);
            copyViews();
            
        }

    }).catch(function (error) {

        console.log();
        console.log("   ---------------------------------------------------------------");
        console.log("    ERROR in updating view");
        console.log("   ---------------------------------------------------------------");
        console.log();
        console.log(error);
        console.log();
        console.log();

    });
    
}